package l2ft.database;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Hashtable;
import java.util.logging.Level;
import java.util.logging.Logger;

import l2ft.Config;
import l2ft.Server;
import l2ft.util.SqlBatch;

import com.mchange.v2.c3p0.ComboPooledDataSource;
import com.mchange.v2.c3p0.DataSources;

public class L2DatabaseFactory
{
	private ComboPooledDataSource _source;
	public final int max_allowed_packet;
	private final Hashtable<String, ThreadConnection> Connections = new Hashtable<String, ThreadConnection>();

	static Logger _log = Logger.getLogger(L2DatabaseFactory.class.getName());

	public L2DatabaseFactory(String url, String login, String pass, int poolSize, int idleTimeOut) throws SQLException
	{
		int _max_allowed_packet = 0;
		try
		{
			if (Config.DATABASE_MAX_CONNECTIONS < 2)
			{
				Config.DATABASE_MAX_CONNECTIONS = 2;
				_log.warning("at least " + Config.DATABASE_MAX_CONNECTIONS + " db connections are required.");
			}

			Class.forName(Config.DATABASE_DRIVER).newInstance();

			_source = new ComboPooledDataSource();
			_source.setDriverClass(Config.DATABASE_DRIVER);
			_source.setJdbcUrl(url);
			_source.setUser(login);
			_source.setPassword(pass);
			_source.setAutoCommitOnClose(true);
			_source.setInitialPoolSize(1);
			_source.setMinPoolSize(1);
			_source.setMaxPoolSize(poolSize);
			_source.setAcquireRetryAttempts(0);
			_source.setAcquireRetryDelay(100);
			_source.setCheckoutTimeout(0);
			_source.setAcquireIncrement(5);
			_source.setMaxStatements(100);
			_source.setIdleConnectionTestPeriod(Config.DATABASE_IDLE_TEST_PERIOD);
			_source.setMaxIdleTime(idleTimeOut);
			_source.setNumHelperThreads(5);
			_source.setBreakAfterAcquireFailure(false);

			_source.getConnection().close();

			ThreadConnection con = null;
			FiltredPreparedStatement statement = null;
			ResultSet rs = null;
			try
			{
				con = getConnection();
				statement = con.prepareStatement("SHOW VARIABLES LIKE 'max_allowed_packet'");
				rs = statement.executeQuery();
				rs.next();
				_max_allowed_packet = rs.getInt(2);
			}
			catch (Exception e)
			{
				e.printStackTrace();
				Server.exit(0, e.getMessage());
			}
			finally
			{
				DatabaseUtils.closeDatabaseCSR(con, statement, rs);
			}

		}
		catch (SQLException x)
		{
			throw x;
		}
		catch (Exception e)
		{
			throw new SQLException("could not init DB connection:" + e);
		}
		max_allowed_packet = _max_allowed_packet;
		SqlBatch.setDefaultLimit(max_allowed_packet);
	}

	public static L2DatabaseFactory getInstance()
	{
		return Config.DATABASE.instance;
	}

	public ThreadConnection getConnection() throws SQLException
	{
		ThreadConnection connection;
		if (Config.USE_DATABASE_LAYER)
		{
			String key = generateKey();

			connection = Connections.get(key);
			if (connection == null)
			{
				try
				{
					connection = new ThreadConnection(_source.getConnection(), this);
				}
				catch (SQLException e)
				{
					_log.warning("Couldn't create connection. Cause: " + e.getMessage());
				}
			}
			else
			{
				connection.updateCounter();
			}

			if (connection != null) synchronized (Connections)
			{
				Connections.put(key, connection);
			}
		}
		else
		{
			connection = new ThreadConnection(_source.getConnection(), this);
		}
		return connection;
	}

	public ThreadConnection getNewConnection() throws SQLException
	{
		return new ThreadConnection(_source.getConnection(), this);
	}

	public int getBusyConnectionCount() throws SQLException
	{
		return _source.getNumBusyConnectionsDefaultUser();
	}

	public int getIdleConnectionCount() throws SQLException
	{
		return _source.getNumIdleConnectionsDefaultUser();
	}

	public Hashtable<String, ThreadConnection> getConnections()
	{
		return Connections;
	}

	public void shutdown()
	{
		_source.close();
		Connections.clear();
		try
		{
			DataSources.destroy(_source);
		}
		catch (SQLException e)
		{
			_log.log(Level.INFO, "", e);
		}
	}

	public String generateKey()
	{
		return String.valueOf(Thread.currentThread().hashCode());
	}
}