package l2ft.gameserver.model.entity.olympiad;

import java.util.concurrent.ScheduledFuture;

import l2ft.Config;
import l2ft.common.ThreadPoolManager;
import l2ft.gameserver.cache.Msg;
import l2ft.gameserver.serverpackets.SystemMessage;
import l2ft.util.Log;

public class OlympiadGameTask implements Runnable
{
	private OlympiadGame _game;
	private BattleStatus _status;
	private int _count;
	private long _time;
	private boolean _terminated = false;

	public boolean isTerminated()
	{
		return _terminated;
	}

	public BattleStatus getStatus()
	{
		return _status;
	}

	public int getCount()
	{
		return _count;
	}

	public OlympiadGame getGame()
	{
		return _game;
	}

	public long getTime()
	{
		return _count;
	}

	public ScheduledFuture<?> shedule()
	{
		return ThreadPoolManager.getInstance().scheduleGeneral(this, _time);
	}

	public OlympiadGameTask(OlympiadGame game, BattleStatus status, int count, long time)
	{
		_game = game;
		_status = status;
		_count = count;
		_time = time;
	}

	@Override
	public void run()
	{
		if(_game == null || _terminated)
			return;

		OlympiadGameTask task = null;

		int gameId = _game.getGameId();

		try
		{
			if (!Olympiad.inCompPeriod())
				return;

			// Прерываем игру, если один из игроков не онлайн, и игра еще не прервана
			if(!_game.checkPlayersOnline() && _status != BattleStatus.ValidateWinner && _status != BattleStatus.Ending)
			{
				Log.add("Player is offline for game " + gameId + ", status: " + _status, "olympiad");
				_game.endGame(1000, true);
				return;
			}

			switch (_status)
			{
				case Begining:
				{
					int teleportToArenaTime = Config.ALT_OLY_TELEPORT_TO_ARENA_TIME;
					_game.broadcastPacket(new SystemMessage(1492).addNumber(teleportToArenaTime), true, false);
					int nextTime = teleportToArenaTime < 5 ? 5 : 15 < teleportToArenaTime ? 15 : 30 < teleportToArenaTime ? 30 : 60 < teleportToArenaTime ? 60 : 120;
					teleportToArenaTime -= nextTime;
					task = new OlympiadGameTask(_game, BattleStatus.Begin_Countdown, nextTime, teleportToArenaTime * 1000);
					break;
				}
				case Begin_Countdown:
				{
					_game.broadcastPacket(new SystemMessage(1492).addNumber(_count), true, false);
					if(_count == 60)
						task = new OlympiadGameTask(_game, BattleStatus.Begin_Countdown, 30, 30000);
					else if(_count == 30)
						task = new OlympiadGameTask(_game, BattleStatus.Begin_Countdown, 15, 15000);
					else if(_count == 15)
						task = new OlympiadGameTask(_game, BattleStatus.Begin_Countdown, 5, 10000);
					else if(_count < 6 && _count > 1)
						task = new OlympiadGameTask(_game, BattleStatus.Begin_Countdown, _count - 1, 1000);
					else if(_count == 1)
						task = new OlympiadGameTask(_game, BattleStatus.PortPlayers, 0, 1000);
					break;
				}
				case PortPlayers:
				{
					_game.portPlayersToArena();
					_game.managerShout();
					task = new OlympiadGameTask(_game, BattleStatus.Started, Config.ALT_OLY_START_BATTLE_TIME, 1000);
					break;
				}
				case Started:
				{
					if(_count == Config.ALT_OLY_START_BATTLE_TIME)
					{
						_game.setState(1);
						_game.preparePlayers();
						if(_game.getType().hasBuffer())
						{
							_game.showBuffers();
						}
					}
					if(_count == 50)
					{
						_game.restoreHpMpCp();
					}
					_game.broadcastPacket(new SystemMessage(1495).addNumber(_count), true, true);
					_count -= 10;

					if(_count > 0)
					{
						task = new OlympiadGameTask(_game, BattleStatus.Started, _count, 10000);
						break;
					}
					_game.hideBuffers();
					_game.openDoors();

					task = new OlympiadGameTask(_game, BattleStatus.CountDown, 5, 5000);
					break;
				}
				case CountDown:
				{
					_game.broadcastPacket(new SystemMessage(1495).addNumber(_count), true, true);
					_count--;
					if(_count <= 0)
						task = new OlympiadGameTask(_game, BattleStatus.StartComp, 36, 1000);
					else
						task = new OlympiadGameTask(_game, BattleStatus.CountDown, _count, 1000);
					break;
				}
				case StartComp:
				{
					if(_count == 36)
					{
						_game.setState(2);
						_game.broadcastPacket(Msg.STARTS_THE_GAME, true, true);
						_game.broadcastInfo(null, null, false);
						_game.getTeam1().processStartQuestEvent();
						_game.getTeam2().processStartQuestEvent();
					}
					// Wait 3 mins (Battle)
					_count--;
					if(_count == 0)
						task = new OlympiadGameTask(_game, BattleStatus.ValidateWinner, 0, 10000);
					else
						task = new OlympiadGameTask(_game, BattleStatus.StartComp, _count, 10000);
					break;
				}
				case ValidateWinner:
				{
					try
					{
						_game.validateWinner(_count > 0);
					}
					catch(Exception e)
					{
						e.printStackTrace();
					}
					task = new OlympiadGameTask(_game, BattleStatus.Ending, 0, 20000);
					break;
				}
				case Ending:
				{
					_game.portPlayersBack();
					_game.clearSpectators();
					_terminated = true;
					if(Olympiad._manager != null)
						Olympiad._manager.freeOlympiadInstance(_game);
					return;
				}
			}
			if(task == null)
			{
				Log.add("task == null for game " + gameId, "olympiad");
				Thread.dumpStack();
				_game.endGame(1000, true);
				return;
			}

			_game.sheduleTask(task);
		}
		catch(Exception e)
		{
			Log.add("Error for game " + gameId + " :" + e.getMessage(), "olympiad");
			e.printStackTrace();
			_game.endGame(1000, true);
		}
	}
}