package l2ft.gameserver.model.entity.olympiad;

import java.util.List;
import java.util.ArrayList;

import l2ft.Config;
import l2ft.gameserver.cache.Msg;
import l2ft.gameserver.model.L2Clan;
import l2ft.gameserver.model.L2ObjectsStorage;
import l2ft.gameserver.model.L2Party;
import l2ft.gameserver.model.L2Player;
import l2ft.gameserver.model.items.L2ItemInstance;
import l2ft.gameserver.model.quest.QuestState;
import l2ft.gameserver.serverpackets.ExOlympiadMatchResult;
import l2ft.gameserver.serverpackets.ExOlympiadMatchResult.MatchResultMemberInfo;
import l2ft.gameserver.serverpackets.ExOlympiadUserInfo;
import l2ft.gameserver.serverpackets.L2GameServerPacket;
import l2ft.gameserver.serverpackets.SystemMessage;
import l2ft.gameserver.skills.EffectType;
import l2ft.util.GArray;
import l2ft.util.GCSArray;
import l2ft.util.Log;

public class OlympiadTeam
{
	private OlympiadGame _game;
	private GCSArray<TeamMember> _members;
	private String _name = "";
	private int _side;
	private double _damage;

	public OlympiadTeam(OlympiadGame game, int side)
	{
		_game = game;
		_side = side;
		_members = new GCSArray<TeamMember>();
	}

	public void addMember(int obj_id)
	{
		String player_name = "";
		L2Player player = L2ObjectsStorage.getPlayer(obj_id);
		if (player != null)
			player_name = player.getName();
		else
		{
			NobleInfo noble = Olympiad._nobles.get(obj_id);
			if (noble != null)
			{
				player_name = noble.char_name;
			}
		}
		_members.add(new TeamMember(obj_id, player_name, _game, _side));

		switch (_game.getType())
		{
			case CLASSED:
			case NON_CLASSED:
				_name = player_name;
				break;
			case TEAM_RANDOM:
				_name = ("Team " + _side);
				break;
			case TEAM:
				if (_name.isEmpty()) break;
					_name = (player_name + " team");
		}
	}

	public void addDamage(double damage)
	{
		_damage += damage;
	}

	public double getDamage()
	{
		return _damage;
	}

	public String getName()
	{
		return _name;
	}

	public void portPlayersToArena()
	{
		for (TeamMember member : _members)
			member.portPlayerToArena();
	}

	public void portPlayersBack()
	{
		for (TeamMember member : _members)
			member.portPlayerBack();
	}

	public void preparePlayers()
	{
		for (TeamMember member : _members)
		{
			member.preparePlayer();
		}
		if (_members.size() <= 1)
			return;

		GArray<L2Player> list = new GArray<L2Player>();

		for (TeamMember member : _members)
		{
			L2Player player;
			if ((player = member.getPlayer()) != null)
			{
				list.add(player);
				L2Party party;
				if ((party = player.getParty()) != null)
				{
					party.oustPartyMember(player);
				}
			}
		}
		if (list.size() <= 1)
			return;

		L2Player leader = list.get(0);
		if (leader == null)
			return;

		L2Party party = new L2Party(leader, 0);
		for (L2Player player_ : list)
			if (player_ != leader)
				player_.joinParty(party);
	}

	public void restoreHpMpCp()
	{
		for (TeamMember member : _members)
			member.restoreHpMpCp();
	}

	public void takePointsForCrash()
	{
		for (TeamMember member : _members)
			member.takePointsForCrash();
	}

	public boolean checkPlayers()
	{
		for (TeamMember member : _members)
			if (member.checkPlayer())
				return true;
		return false;
	}

	public boolean isAllDead()
	{
		for (TeamMember member : _members)
			if ((!member.isDead()) && (member.checkPlayer()))
				return false;
		return true;
	}

	public boolean contains(int objId)
	{
		for (TeamMember member : _members)
			if (member.getObjId() == objId)
				return true;
		return false;
	}

	public GArray<L2Player> getPlayers()
	{
		GArray<L2Player> players = new GArray<L2Player>();

		for (TeamMember member : _members)
		{
			L2Player player = member.getPlayer();
			if (player != null)
			players.add(player);
		}
		return players;
	}

	public GCSArray<TeamMember> getMembers()
	{
		return _members;
	}

	public void broadcast(L2GameServerPacket p)
	{
		for (TeamMember member : _members)
		{
			L2Player player = member.getPlayer();
			if (player != null)
				player.sendPacket(new L2GameServerPacket[] { p });
		}
	}

	public void broadcastInfo()
	{
		for (TeamMember member : _members)
		{
			L2Player player = member.getPlayer();
			if (player != null)
				player.broadcastPacket(new ExOlympiadUserInfo(player, player.getOlympiadSide()));
		}
	}

	public boolean logout(L2Player player)
	{
		if (player != null)
			for (TeamMember member : _members)
			{
				L2Player pl = member.getPlayer();
				if ((pl != null) && (pl == player))
					member.logout();
			}
		return checkPlayers();
	}

	public boolean doDie(L2Player player)
	{
		if (player != null)
			for (TeamMember member : _members)
			{
				L2Player pl = member.getPlayer();
				if ((pl != null) && (pl == player))
					member.doDie();
			}
		return isAllDead();
	}

	public void winGame(OlympiadTeam looseTeam)
	{
		int totalPointDiff = 0; int curPointDiff = 0;

		List<MatchResultMemberInfo> team1info = new ArrayList<MatchResultMemberInfo>(_members.size());
		List<MatchResultMemberInfo> team2info = new ArrayList<MatchResultMemberInfo>(_members.size());
		GCSArray<TeamMember> looseTeamMembers = looseTeam.getMembers();

		for (int i = 0; i < _members.size(); i++) 
		{
			try
			{
				TeamMember member = _members.get(i);
				TeamMember lose_member = looseTeamMembers.get(i);
				curPointDiff = transferPoints(lose_member.getStat(), member.getStat());
				totalPointDiff += curPointDiff;

				String name = member.getName();
				L2Clan clan = member.getClan();
				String clanName = clan == null ? "" : clan.getName();
				int clanId = clan == null ? 0 : clan.getClanId();
				int classId = member.getStat().class_id;
				int points = member.getStat().olympiad_points;
				team1info.add(new ExOlympiadMatchResult.MatchResultMemberInfo(name, clanName, clanId, classId, (int)getDamage(), points, curPointDiff));

				name = lose_member.getName();
				clan = lose_member.getClan();
				clanName = clan == null ? "" : clan.getName();
				clanId = clan == null ? 0 : clan.getClanId();
				classId = lose_member.getStat().class_id;
				points = lose_member.getStat().olympiad_points;
				team2info.add(new ExOlympiadMatchResult.MatchResultMemberInfo(name, clanName, clanId, classId, (int)looseTeam.getDamage(), points, -curPointDiff));
			}
			catch (Exception e)
			{
				e.printStackTrace();
			}
		}
		for (L2Player player : getPlayers())
		{
			try
			{
				L2ItemInstance item = player.getInventory().addItem(Config.ALT_OLY_BATTLE_REWARD_ITEM, _game.getType().getReward());
				player.sendPacket(SystemMessage.obtainItems(item.getItemId(), _game.getType().getReward(), 0));
			}
			catch (Exception e)
			{
				e.printStackTrace();
			}
		}

		for (L2Player player : _game.getAllPlayers())
		{
			player.getEffectList().stopEffects(EffectType.Vitality);
		}
		_game.broadcastPacket(new ExOlympiadMatchResult(getName(), team1info, team2info), true, false);
		_game.broadcastPacket(new SystemMessage(1497).addString(getName()), true, true);
		_game.broadcastPacket(new SystemMessage(1657).addString(getName()).addNumber(Integer.valueOf(totalPointDiff)), true, false);
		_game.broadcastPacket(new SystemMessage(1658).addString(looseTeam.getName()).addNumber(Integer.valueOf(totalPointDiff)), true, false);

		processEndQuestEvent(1);
		looseTeam.processEndQuestEvent(0);

		Log.add("Olympiad Result: " + getName() + " vs " + looseTeam.getName() + " ... (" + (int)_damage + " vs " + (int)looseTeam.getDamage() + ") " + getName() + " win " + totalPointDiff + " points", "olympiad");
	}

	public void tie(OlympiadTeam otherTeam)
	{
		for (int i = 0; i < _members.size(); i++)
		{
			try
			{
				NobleInfo stat1 = (getMembers().get(i)).getStat();
				NobleInfo stat2 = (otherTeam.getMembers().get(i)).getStat();
				stat1.olympiad_points -= 2;
				stat2.olympiad_points -= 2;
			}
			catch (Exception e)
			{
				e.printStackTrace();
			}
		}
		_game.broadcastPacket(Msg.THE_GAME_ENDED_IN_A_TIE, true, true);

		processEndQuestEvent(2);
		otherTeam.processEndQuestEvent(2);

		Log.add("Olympiad Result: " + getName() + " vs " + otherTeam.getName() + " ... tie", "olympiad");
	}

	private int transferPoints(NobleInfo from, NobleInfo to)
	{
		int fromPoints = from.olympiad_points;
		int toPoints = to.olympiad_points;

		int pointDiff = Math.max(1, Math.min(fromPoints, toPoints) / _game.getType().getLooseMult());
		pointDiff = pointDiff > 10 ? 10 : pointDiff;

		from.olympiad_points = (fromPoints - pointDiff);
		from.competitions_loose += 1;
		from.competitions_done += 1;

		to.olympiad_points = (toPoints + pointDiff);
		to.competitions_win += 1;
		to.competitions_done += 1;

		switch (_game.getType().ordinal())
		{
			case 1:
				from.competitions_week_0 += 1;
				to.competitions_week_0 += 1;
				break;
			case 2:
				from.competitions_week_1 += 1;
				to.competitions_week_1 += 1;
				break;
			case 3:
				from.competitions_week_2 += 1;
				to.competitions_week_2 += 1;
				break;
			case 4:
				from.competitions_week_3 += 1;
				to.competitions_week_3 += 1;
		}

		return pointDiff;
	}

	public void saveNobleData()
	{
		for (TeamMember member : _members)
			member.saveNobleData();
	}

	public void processStartQuestEvent()
	{
		for (L2Player player : getPlayers())
			try
			{
				for (QuestState qs : player.getAllQuestsStates())
					if (qs != null && qs.isStarted() && (qs.getQuest() instanceof QuestEvent))
						((QuestEvent)qs.getQuest()).startMatch(player, qs, _side);
			}
			catch (Exception e)
			{
				e.printStackTrace();
			}
	}

  	public void processEndQuestEvent(int win)
	{
		for (L2Player player : getPlayers())
			try
			{
				for (QuestState qs : player.getAllQuestsStates())
				if (qs != null && qs.isStarted() && (qs.getQuest() instanceof QuestEvent))
					((QuestEvent)qs.getQuest()).endMatch(player, qs, win);
			}
			catch (Exception e)
			{
				e.printStackTrace();
			}
	}
}